/*
 * periodicstats.c - periodic stats method handeling code
 *
 * Copyright (C) 2022 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Shubham Sharma <shubham.sharma@iopsys.eu>
 *
 * See LICENSE file for license related information.
 *
 */

#include <time.h>
#include <libubus.h>

#include "periodicstats.h"

#define ubus_timeout 5

static void _ubus_handler_cb(struct ubus_request *req, int type __attribute__((unused)), struct blob_attr *msg)
{
	struct json_object **jobj = (struct json_object **)req->priv;
	char *buffer = NULL;

	if (msg == NULL || jobj == NULL)
		return;

	*jobj = NULL;

	buffer = blobmsg_format_json(msg, true);
	if (buffer == NULL)
		return;

	*jobj = json_tokener_parse(buffer);

	free(buffer);
}

static int _ubus_call(const char *obj_name, const char *method, struct blob_buf *bb, struct json_object **jout)
{
	struct ubus_context *ubus_ctx = NULL;
	uint32_t id;
	int ret = -1;

	ubus_ctx = ubus_connect(NULL);
	if (ubus_ctx == NULL) {
		syslog(LOG_ERR, "ubus connect failed");
		return -1;
	}

	ret = ubus_lookup_id(ubus_ctx, obj_name, &id);
	if (ret != 0) {
		syslog(LOG_ERR, "ubus lookup failed");
		return -1;
	}

	ret = ubus_invoke(ubus_ctx, id, method, bb->head, _ubus_handler_cb, jout, ubus_timeout * 1000);
	ubus_free(ubus_ctx);

	return ret;
}

static void usp_get_parameter_value(const char *path_name, char **value, char **type)
{
	struct blob_buf bb = {0};
	json_object *jobj = NULL;

	if (value == NULL || type == NULL)
		return;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", path_name);
	void *table = blobmsg_open_table(&bb, "optional");
	blobmsg_add_string(&bb, "format", "raw");
	blobmsg_close_table(&bb, table);

	_ubus_call("bbfdm", "get", &bb, &jobj);
	if (jobj == NULL) {
		syslog(LOG_ERR, "Could not read stats");
		goto end;
	}

	json_object *res = NULL;
	json_object_object_get_ex(jobj, "results", &res);

	if (res == NULL || json_object_array_length(res) != 1) {
		syslog(LOG_ERR, "BBFDM get response invalid");
		goto end;
	}

	json_object *res_ob = json_object_array_get_idx(res, 0);
	json_object_object_foreach(res_ob, key, val) {
		if (key == NULL)
			continue;

		if (strcmp(key, "data") == 0) {
			*value = strdup(json_object_get_string(val));
			continue;
		}

		if (strcmp(key, "type") == 0) {
			*type = strdup(json_object_get_string(val));
			continue;
		}
	}

	json_object_put(jobj);

end:
	blob_buf_free(&bb);
	return;
}

static char *uci_get_option_value(const char *sampleset, const char *reference, const char *option)
{
	struct uci_context *uci_ctx = uci_alloc_context();
	struct uci_package *uci_pkg = NULL;
	struct uci_element *uci_elmnt = NULL;
	char *value = NULL;

	uci_load(uci_ctx, "periodicstats", &uci_pkg);
	if (!uci_pkg) {
		syslog(LOG_ERR, "Failed to load configuration\n");
		uci_free_context(uci_ctx);
		return value;
	}

	uci_foreach_element(&uci_pkg->sections, uci_elmnt) {
		struct uci_section *uci_sec = uci_to_section(uci_elmnt);
		if (uci_sec && !strcmp(uci_sec->type, "parameter")) {
			struct uci_option *samp_set = uci_lookup_option(uci_ctx, uci_sec, "sample_set");
			struct uci_option *ref = uci_lookup_option(uci_ctx, uci_sec, "reference");
			if (samp_set && ref && !strcmp(samp_set->v.string, sampleset) && !strcmp(ref->v.string, reference)) {
				struct uci_option *u_option = uci_lookup_option(uci_ctx, uci_sec, option);
				value = u_option ? strdup(u_option->v.string) : NULL;
				break;
			}
		}
	}

	uci_unload(uci_ctx, uci_pkg);
	uci_free_context(uci_ctx);

	return value;
}

void uci_set_option_value(const char *sec_name, const char *option, const char *value)
{
	struct uci_context *uci_ctx = uci_alloc_context();
	struct uci_ptr ptr = {0};
	char uci_path[256] = {0};

	snprintf(uci_path, sizeof(uci_path), "periodicstats.%s.%s=%s", sec_name, option, value);

	if (uci_lookup_ptr(uci_ctx, &ptr, uci_path, true) != UCI_OK) {
		uci_free_context(uci_ctx);
		return;
	}

	if (uci_set(uci_ctx, &ptr) != UCI_OK) {
		uci_free_context(uci_ctx);
		return;
	}

	if (uci_save(uci_ctx, ptr.p) != UCI_OK) {
		uci_free_context(uci_ctx);
		return;
	}

	uci_commit(uci_ctx, &ptr.p, false);
	uci_free_context(uci_ctx);
}

static time_t get_next_sample_time(time_t time_ref, uint32_t periodic_interval)
{
	unsigned int next_period = 0;
	time_t curr_time;

	curr_time = time(NULL);

	if (curr_time > time_ref)
		next_period = (periodic_interval - ((curr_time - time_ref) % periodic_interval));
	else
		next_period = (time_ref - curr_time) % periodic_interval;

	if (next_period == 0)
		next_period = periodic_interval;

	return (curr_time + next_period);
}

static unsigned int pstats_get_nb_sample_stored(char *list)
{
	unsigned int count = 0;

	if (!list)
		return count;

	if (*list)
		count = 1;

	while (*list) {
		if (*list++ == ',')
			count++;
	}

	return count;
}

static bool data_type_numeric(char *type)
{
	bool ret = false;

	if (type == NULL)
		return ret;

	if (strcmp(type, "xsd:unsignedInt") == 0 ||
	    strcmp(type, "xsd:int") == 0 ||
	    strcmp(type, "xsd:unsignedLong") == 0 ||
	    strcmp(type, "xsd:long") == 0) {
		ret = true;
	}

	return ret;
}

static void reset_sampling_params(param *sample_param)
{
	int i;

	if (sample_param == NULL)
		return;

	sample_param->param_numeric = false;
	for (i = 0; i < INTERNAL_SAMPLE_RATE; i++) {
		FREE(sample_param->values[i]);
	}
}

static void generate_report(smpl_set *sample_set, param *sample_param)
{
	int mode, i, avg_devider = 0;
	char *report_val = NULL;
	long total_val = 0, avg_val = 0;
	char str_val[64] = {0};

	if (sample_set == NULL || sample_param == NULL) {
		goto end;
	}

	if (sample_param->param_numeric == false) {
		mode = CAL_MODE_LATEST;
	} else {
		mode = sample_param->calculation_mode;
	}

	switch (mode) {
	case CAL_MODE_AVERAGE:
		for (i = 0; i < INTERNAL_SAMPLE_RATE; i++) {
			if (sample_param->values[i] == NULL) {
				continue;
			}
			avg_devider++;
			total_val = total_val + strtol(sample_param->values[i], NULL, 10);
		}

		if (avg_devider == 0) {
			break;
		}

		avg_val = total_val/avg_devider;
		snprintf(str_val, sizeof(str_val), "%ld", avg_val);
		report_val = str_val;
		break;
	case CAL_MODE_MINIMUM:
		for (i = 0; i < INTERNAL_SAMPLE_RATE; i++) {
			if (sample_param->values[i] == NULL) {
				continue;
			}

			if (report_val == NULL) {
				report_val = sample_param->values[i];
			} else {
				if (strtol(report_val, NULL, 10) > strtol(sample_param->values[i], NULL, 10)) {
					report_val = sample_param->values[i];
				}
			}
		}
		break;
	case CAL_MODE_MAXIMUM:
		for (i = 0; i < INTERNAL_SAMPLE_RATE; i++) {
			if (sample_param->values[i] == NULL) {
				continue;
			}

			if (report_val == NULL) {
				report_val = sample_param->values[i];
			} else {
				if (strtol(report_val, NULL, 10) < strtol(sample_param->values[i], NULL, 10)) {
					report_val = sample_param->values[i];
				}
			}
		}
		break;
	case CAL_MODE_LATEST:
		i = INTERNAL_SAMPLE_RATE - 1;
		report_val = sample_param->values[i];
		break;
	}

	if (report_val != NULL) {
		char *values = NULL;
		char *sample_list = uci_get_option_value(sample_set->name, sample_param->reference, "values");

		if (sample_list) {
			unsigned int samples_stored = pstats_get_nb_sample_stored(sample_list);
			if (samples_stored >= sample_set->rprt_smpls) {
				char *tmp = strchr(sample_list, ',') + 1;
				int buf_size = strlen(tmp) + strlen(report_val) + 2;

				values = (char *)calloc(buf_size, sizeof(char));
				if (values != NULL) {
					snprintf(values, buf_size, "%s,%s", tmp, report_val);
				}
			} else {
				int buf_size = strlen(sample_list) + strlen(report_val) + 2;

				values = (char *)calloc(buf_size, sizeof(char));
				if (values != NULL) {
					snprintf(values, buf_size, "%s,%s", sample_list, report_val);
				}
			}

			free(sample_list);
		} else {
			int buf_size = strlen(report_val) + 1;

			values = (char *)calloc(buf_size, sizeof(char));
			if (values != NULL) {
				snprintf(values, buf_size, "%s", report_val);
			}
		}

		if (values != NULL) {
			uci_set_option_value(sample_param->name, "values", values);
			free(values);
		}
	}

end:
	reset_sampling_params(sample_param);
	return;
}

static void pstats_handle_sampleset(smpl_set *sample_set, param *sample_param)
{
	char *data = NULL, *type = NULL;

	if (sample_set == NULL || sample_param == NULL) {
		return;
	}

	usp_get_parameter_value(sample_param->reference, &data, &type);
	if (data == NULL || type == NULL) {
		syslog(LOG_ERR, "Value and/or Type is NULL");
	} else {
		sample_param->values[sample_set->int_smpl_count] = strdup(data);
		sample_param->param_numeric = data_type_numeric(type);
	}

	FREE(data);
	FREE(type);

	if (sample_set->int_smpl_count == INTERNAL_SAMPLE_RATE - 1) { // sample count start with 0
		generate_report(sample_set, sample_param);
	}

	return;
}

static void pstats_sampleset_cb(struct uloop_timeout *timeout)
{
	smpl_set *sample_set = NULL;
	param *sample_param = NULL;
	unsigned long next_internal_time;

	sample_set = container_of(timeout, smpl_set, utimer);

	list_for_each_entry(sample_param, &sample_set->params, list) {
		if ((sample_param->enable != 1) || (sample_param->reference[0] == '\0'))
			continue;

		pstats_handle_sampleset(sample_set, sample_param);
	}

	sample_set->int_smpl_count++;

	if (sample_set->int_smpl_count != INTERNAL_SAMPLE_RATE) {
		time_t cur_time = time(NULL);
		if (sample_set->next_sample_time <= cur_time) {
			next_internal_time = 10; // start immediate
		} else {
			unsigned long time_remaining;
			time_remaining = (sample_set->next_sample_time - cur_time) * 1000;
			next_internal_time = time_remaining / (INTERNAL_SAMPLE_RATE - sample_set->int_smpl_count);
		}
	} else {
		sample_set->int_smpl_count = 0;
		sample_set->next_sample_time = get_next_sample_time(sample_set->time_ref, sample_set->smpl_intrvl);
		next_internal_time = ((sample_set->next_sample_time - time(NULL)) * 1000) / INTERNAL_SAMPLE_RATE;
	}

	syslog(LOG_INFO, "Start new collection session of sample set '%s' in %lu msec\n", sample_set->name, next_internal_time);
	uloop_timeout_set(timeout, next_internal_time);
}

/**
 *  pstats_get_stats function for invoking periodic statistics
 *  @param global_list input pointer to global list_head structure
 */
void pstats_get_stats(struct list_head *global_list)
{
	smpl_set *sample_set = NULL;
	unsigned long next_internal_time;

	list_for_each_entry(sample_set, global_list, list) {
		if (sample_set->enable != 1)
			continue;

		sample_set->utimer.cb = pstats_sampleset_cb;
		sample_set->next_sample_time = get_next_sample_time(sample_set->time_ref, sample_set->smpl_intrvl);
		next_internal_time = ((sample_set->next_sample_time - time(NULL)) * 1000) / INTERNAL_SAMPLE_RATE;
		syslog(LOG_INFO, "Start new collection session of sample set '%s' in %lu msec\n", sample_set->name, next_internal_time);
		uloop_timeout_set(&sample_set->utimer, next_internal_time);
	}
}
